# project
class Config:
    """
     the structure of self.config file
    """

    STEP_CONFIG = "step_config"
    ARGS = "args"
    PROJECT_CONFIG = "project_config"
    STEP_NAME = "step_name"
    CLIENT = "client"
    WORKFLOW = "workflow"
    USER = "user"
    FRAMEWORK = "framework"
    ENV = "env"
    JOB_ID = "job_id"
    LOG_LEVEL = "log_level"
    SSD_JOB_ID = "ssd_job_id"
    AWS_KEY = "awskey"
    AWS_SECRET = "awssecret"
    VAULT_KEY = "vault_key"
    VAULT_METADATA = "vault_metadata"
    PROJECT_TEST = "project_test.yaml"
    DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"
    DATETIME_FORMAT_FILENAME = "%Y%m%dT%H%M%S "
    FISCAL_MONTH_END = "fiscal_month_end"
    CURRENT_YEAR = "current_year"
    PROJECT_ROOT_PATH = "project_root_path"
    CLIENT2 = "client2"
    CMS_LOCATION = "cms_location"
    NPI_PARQUET_LOCATION = "npi_parquet_location"
    LOCATION_MASTER = "master_location"
    LOCATION_SSD = "ssd_location"
    LOCATION_AGGR = "aggr_location"
    LOCATION_REFERENCE = "reference_location"
    LOCATION_REPORT = "report_location"
    PARENT_DEMO_CLIENT = "parent_demo_client"

## project_test
class ProjectTest:
    AWS_TEST_DIRECTORY = "aws_test_directory"
    AWS_TEST_DIRECTORY_BUCKET = "bucket"
    AWS_TEST_DIRECTORY_BASE_KEY = "base_key"

    TEST_DATA_FACTORY = "test_data_factory"
    TEST_DATA_FACTORY_FRACTION = "fraction"
    TEST_DATA_FACTORY_LIMIT = "limit"

    TEST_ETL_JOB = "test_etl_job"
    TEST_ETL_JOB_JOBS = "jobs"



"""
unit testing handler
~~~~~~~
This module contains a function to take sample test_data from input and output of functions

The location to store test data is in AWS.
"""

import json
from dependencies.aws import cadmium_aws
from constants.constants_configs import ProjectTest, Config as Constant_Config
from datetime import datetime


def unit_test_data_factory(func):
    '''
    A decorator to generate sample data for functions.
    Assumptions about the function:
        1st argument is self
        2nd argument is a dataframe  (usually represents a large data)
        3rd and onwards are other config parameters (no large data)

    The decorator gets a fraction of input data (fraction is set in project_test config file) and calculates
    the expected output for that fraction. It stores that data into an AWS bucket (bucket and key are in configuration
    file)
    bucket and key are stored in project_test.yaml:
    aws_test_directory:
          bucket: "<BUCKET>"
          base_key: "<KEY>"
    There are three files that are stored in bucket/base_key folder:
    files are named:
        - test_data-<CLASS_NAME>-<METHOD_NAME>-args_<DATETIME>.json
            This file includes the config, vault and arguments (ARGS[2:])
        - test_data-<CLASS_NAME>-<METHOD_NAME>-input_<DATETIME>.pickle
            This file includes the sample of input dataframe (ARGS[1])
        - test_data-<CLASS_NAME>-<METHOD_NAME>-output_<DATETIME>.pickle
            This file includes the output for the sampled input

    The decorator then runs the function as normal.

    How to use it:
        1- upon completing or editing a  function and when you are confident that the function is operational and
        bug free, add the decorator label (`@unit_test_data_factory`) to top of the function,
        2- add `self.unittest_sampling_enabled = True` to the main function (before calling the transform)
        3- run the ETL job (function) for a dataset (could be a large dataset). After the job finishes, check for the newly generated sample files in
        aws. (check the logs for  file names).
        4- Remove the decorator label or set `self.unittest_sampling_enabled = False`.
        5- add to unittest conf:    add module_name, class name, function and version(datetime) to `configs/project/project_test.yaml`


    :param func:
    :return:
    '''
    def unittest_inner_function(*args, **kwargs):
        r = func(*args, **kwargs)
        # first argument is the self object
        self = args[0]

        if hasattr(self, "unittest_sampling_enabled"):
            unittest_sampling_enabled = self.unittest_sampling_enabled
        else:
            unittest_sampling_enabled = False

        # unit_test_data_factory configs
        project_test_configs = self.config[Constant_Config.FRAMEWORK][Constant_Config.PROJECT_TEST]
        current_datetime = datetime.now().strftime('%Y%m%d%H%M')
        if unittest_sampling_enabled:

            storage_bucket = project_test_configs[ProjectTest.AWS_TEST_DIRECTORY][ProjectTest.AWS_TEST_DIRECTORY_BUCKET]
            storage_key = project_test_configs[ProjectTest.AWS_TEST_DIRECTORY][ProjectTest.AWS_TEST_DIRECTORY_BASE_KEY]
            storage_aws_path = f"s3a://{storage_bucket}/{storage_key}"

            file_name_args = "{}test_data-{}-{}-args_{}.json".format(
                storage_key, self.__class__.__name__, func.__name__, current_datetime)
            file_path_input = "{}test_data-{}-{}-input_{}.pickle".format(
                storage_aws_path, self.__class__.__name__, func.__name__, current_datetime)
            file_path_output = "{}test_data-{}-{}-output_{}.pickle".format(
                storage_aws_path, self.__class__.__name__, func.__name__, current_datetime)


            # write other configuration arguments like
            args_dict = {
                "config": self.config,
                "vault": self.Vault,
                "args_2_plus": args[2:]
            }
            aws_conn = cadmium_aws()
            aws_conn.write_json_to_bucket(
                json.dumps(args_dict),
                storage_bucket,
                file_name_args
            )
            # args_dict_df = pd.DataFrame.from_dict(args_dict)
            # args_dict_df.to_pickle(file_name_args)

            # sample the input data
            input_spark_df = args[1].sample(
                True,
                project_test_configs[ProjectTest.TEST_DATA_FACTORY][ProjectTest.TEST_DATA_FACTORY_FRACTION],  seed=0
            ).limit(project_test_configs[ProjectTest.TEST_DATA_FACTORY][ProjectTest.TEST_DATA_FACTORY_LIMIT])

            args_2 = (self, input_spark_df) + args[2:]
            # use pandas in order to create one file
            input_spark_df.toPandas().to_pickle(
                file_path_input

            )
            # expected output for sample input
            expected_r = func(*args_2, **kwargs)

            # store expected output
            expected_r.toPandas().to_pickle(
                file_path_output
            )
            self.log.info(f"sample test data created at {file_name_args}, {file_path_input},  {file_path_output}")

        return r
    return unittest_inner_function




from pyspark.sql import SparkSession
from jobs.etl_job_base import EtlJobBase
from dependencies.exception import exception_handler
from py4j.protocol import Py4JJavaError
from constants.constants_configs import Config as Constant_Config
from pyspark.sql.utils import IllegalArgumentException
import urllib.parse
# Requires:
# pyspark --packages org.mongodb.spark:mongo-spark-connector_2.11:2.4.1
# example mongo-match-string: "{$match : {'SourceFileInfo.ISA_Date' : {$lt: '200900', $gt:'200800'}, 'Client_Name' : /thp.*/i}}"

class PullFromMongo(EtlJobBase):

    def get_custom_spark_conf(self):
        port=self.config[Constant_Config.STEP_CONFIG]['port'] if 'port' in self.config[Constant_Config.STEP_CONFIG] else "27017"
        mongo_uri = f"mongodb://{urllib.parse.quote(self.config[Constant_Config.STEP_CONFIG]['username'])}:{urllib.parse.quote(self.config[Constant_Config.STEP_CONFIG]['password'])}@{self.config[Constant_Config.STEP_CONFIG]['host']}:{port}/{self.config[Constant_Config.STEP_CONFIG]['db']}.{self.config[Constant_Config.STEP_CONFIG]['collection']}"
        print("mongo_uri is {}".format(mongo_uri))
        configs = {
            "spark.mongodb.input.uri": mongo_uri
        }

        return configs

    @exception_handler
    def extract_data(self):
        """
        Loads data from specified MongoDB collection according to Mongo Match String
        For more information on how to create the pipeline match string, see:q
        https://docs.mongodb.com/spark-connector/master/python/aggregation/
        https://docs.mongodb.com/manual/core/aggregation-pipeline/
        https://docs.mongodb.com/manual/reference/operator/aggregation/match/#pipe._S_match
        https://docs.mongodb.com/manual/tutorial/query-documents/#read-operations-query-argument
        """


        # Read Mongodb
        try:
            df = (
                self.spark.read.format("mongo")
                .option("samplePoolSize", 100000)
                .option("sampleSize", 100000)
                .option("pipeline", self.config[Constant_Config.STEP_CONFIG]["mongo_match_string"])
                .load(inferSchema=False)
            )
            '''
            TODO: Enhance pull_from_mongo to not infer Schema but to use the schema provided by us. 
            For reference: This can be done by
            df = (
                spark.read.format("mongo")
                    .option("samplePoolSize", 1000000)
                    .option("pipeline", self.config[Constant_Config.STEP_CONFIG]["mongo_match_string"])
                    .load(schema=new_schmea, inferSchema=False) 
            )
            new_schema can be stored using below 
                # This can be do
                    # temp_rdd = self.spark.parallelize(df.schema)
                    # temp_rdd.coalesce(1).saveAsPickleFile("s3a://path/to/destination_schema.pickle")
                    #
                    # df.printSchema()
            '''
            #df.printSchema()

        except Py4JJavaError as e:
            self.log.error(
                "Error connecting to mongo; check credentials and connection"
            )
            self.log.error(e.java_exception.toString())
            raise
        except IllegalArgumentException:
            self.log.error(
                "Invalid Mongo DB Match string; try verifying in Mongo Shell"
            )
            self.log.error(
                self.config[Constant_Config.STEP_CONFIG]["mongo_match_string"]
            )
            raise

        try:
            assert not df.rdd.isEmpty()
        except AssertionError:
            self.log.error("Query returned 0 records; check query:")
            self.log.error(
                self.config[Constant_Config.STEP_CONFIG]["mongo_match_string"]
            )
            raise

        return df

    def transform_data(self, df):
        return df

    @exception_handler
    def load_data(self, df):

        df.write.json(
            f"{self.config[Constant_Config.STEP_CONFIG]['output_location']}",
            mode="append"
            # compression="gzip"
        )
        return

    def get_custom_argparse_configs(self, parser):
        parser.add_argument(
            "--mongo-match-string",
            type=str,
            action="store",
            required=True,
            help="""Mongodb match string to filter data from db; e.g.: "{$match : {'SourceFileInfo.ISA_Date' : {$lt: '200900', $gt:'200800'}}}" """,
        )
        parser.add_argument(
            "--output-location",
            type=str,
            action="store",
            required=True,
            help="Where do you want the data to be written to in s3? Doc Type & Job ID will be added to the path. Do not end in /",
        )

        return parser

    @exception_handler
    def add_custom_config_fields(self, config, args):
        config[Constant_Config.STEP_CONFIG] = {
            "mongo_match_string": args["mongo_match_string"],
            "output_location" : args["output_location"]
        }


        return config


if __name__ == "__main__":
    run_pulL_from_mongo = PullFromMongo()
    run_pulL_from_mongo.main()



import boto3
import json

from botocore.exceptions import ClientError

from dependencies.exception import exception_handler

class cadmium_aws:
    def __init__(self, _awskey=None, _awssecret=None):
        self.bucket_name = ""

    def s3_client(self):
        return boto3.client('s3')

    def s3_resource(self, _awskey=None, _awssecret=None):
        if _awskey is not None:
            return boto3.resource('s3',
                                  aws_access_key_id=_awskey,
                                  aws_secret_access_key=_awssecret)

        return boto3.resource('s3')

    def read_json_from_bucket(self, bucket_name, target_file, _awskey=None, _awssecret=None):
        if _awskey is not None:
            try:
                s3 = self.s3_resource(_awskey=_awskey, _awssecret=_awssecret)
                config_s3 = s3.Object(bucket_name, target_file)
                config_raw = config_s3.get()['Body'].read().decode('utf-8')
                s3_client_json = json.loads(config_raw)
                return s3_client_json

            except ClientError as e:
                print("Unexpected error: %s" % e)
                print(f'No such file {target_file}')
                return 0

        try:
            s3_client_obj = self.s3_client().get_object(Bucket=bucket_name, Key=target_file)
            s3_client_data = s3_client_obj['Body'].read().decode('utf-8')
            s3_client_json = json.loads(s3_client_data)
            return s3_client_json

        except ClientError as e:
            print("Unexpected error: %s" % e)
            print(f'No such file {target_file}')
            return 0

    def write_json_to_bucket(self, json_val,  bucket_name, target_file, _awskey=None, _awssecret=None):
        if _awskey is not None:
            try:
                s3 = self.s3_resource(_awskey=_awskey, _awssecret=_awssecret)
                return s3.Object(
                    bucket_name,
                    target_file
                ).put(
                    Body=json_val.encode('utf-8')
                )

            except ClientError as e:
                print("Unexpected error: %s" % e)
                print(f'No such file {target_file}')
                return 0

        try:


            return self.s3_client().put_object(
                Body=json_val,
                Bucket=bucket_name,
                Key=target_file
            )

        except ClientError as e:
            print("Unexpected error: %s" % e)
            print(f'No such file {target_file}')
            return 0


"""
    This is the Postgres database helpers with connections and other sql related functions
    like insert.
"""

import psycopg2
from dependencies.logger import logger


class CadmiumPostGres:
    _cursor = None

    def __init__(self, host, dbname, user, password, port, auto_commit=False):
        self.create_connection(_host=host, _dbname=dbname, _user=user, _password=password, _port=port,
                               _auto_commit=auto_commit)
        self._log = logger()

    @classmethod
    def create_connection(cls, _host, _dbname, _user, _password, _port, _auto_commit):
        conn = psycopg2.connect(host=_host, dbname=_dbname, user=_user, password=_password, port=_port)
        conn.autocommit = _auto_commit
        cls._cursor = conn.cursor()

    def query_table(self, _query):
        try:
            self._cursor.execute(_query)
            result = self._cursor.fetchone()
            return result
        except (Exception, psycopg2.DatabaseError) as error:
            self._log.error(f"post_gres.py: Query Error : {error}, query: {_query}")

    def query_table_fetchall(self, _query):
        try:
            self._cursor.execute(_query)
            result = self._cursor.fetchall()
            return result
        except (Exception, psycopg2.DatabaseError) as error:
            self._log.error(f"post_gres.py: Query Error : {error}, query: {_query}")

    def execute_table(self, _query):
        try:
            self._cursor.execute(_query)
        except (Exception, psycopg2.DatabaseError) as error:
            self._log.error(f"post_gres.py: Execute Error : {error}, query: {_query}")

    def insert_into_table(self, table_name, columns_tuples, placeholders_tuple, values_tuple):
        query = f"INSERT INTO {table_name} {columns_tuples} VALUES {placeholders_tuple}"
        try:
            self._cursor.execute(query, values_tuple)
        except (Exception, psycopg2.DatabaseError) as error:
            self._log.error(f"post_gres.py: Insert Error : {error}, query: {query}")

    def close(self):
        self._cursor.close()

    @property
    def log(self):
        return self._log



import re
import pyspark.sql.functions as F
from dependencies.exception import exception_handler
from jobs.etl_job_base import EtlJobBase
from constants.constants_configs import Config as Constant_Config

# from dependencies.unit_test_data_factory import unit_test_data_factory


class TextToParquet(EtlJobBase):
    @exception_handler
    def extract_data(self):

        if self.config[Constant_Config.STEP_CONFIG]["filetype"] == "fixed-width":
            data = self.spark.read.text(
                 self.config[Constant_Config.STEP_CONFIG]["input_source"]
            )
        elif self.config[Constant_Config.STEP_CONFIG]["filetype"] == "csv":
            delimiter = self.config[Constant_Config.STEP_CONFIG]["delimiter"] or ","
            data = self.spark.read.csv(
                self.config[Constant_Config.STEP_CONFIG]["input_source"],
                sep=delimiter,
                header=self.config[Constant_Config.STEP_CONFIG]["header_row"],
                ignoreLeadingWhiteSpace=True,
                ignoreTrailingWhiteSpace=True,
            )
        else:
            self.log.error(
                f"Invalid fileyype: {self.config[Constant_Config.STEP_CONFIG]['filetype']}"
            )
            self.log.error(
                "text_to_parquet accepts only 'fixed-width' and 'csv' as valid filetypes."
            )
            raise IOError

        return data

    @exception_handler
    # @unit_test_data_factory
    def transform_data(self, data):

        # Parse fixed-width files
        if self.config[Constant_Config.STEP_CONFIG]["filetype"] == "fixed-width":
            data = self.parse_fixed_width(data)
        # Select & Rename Columns
        keep_cols_dict = {}
        for k, v in self.config[Constant_Config.STEP_CONFIG][
            "input_to_output_cols"
        ].items():
            # Skip metadata columns & columns with no specified output col name
            if (not re.fullmatch(r".*\{.*\}.*", k)) and (v is not None):
                keep_cols_dict[k] = v
        data = data.select(list(keep_cols_dict.keys()))
        for k, v in keep_cols_dict.items():
            data = data.withColumnRenamed(k, v)
        # Add metadata
        data = self.add_metadata(data)

        return data

    @exception_handler
    def load_data(self, output_data):
        if self.config[Constant_Config.STEP_CONFIG]["partition"] is not None:
            output_data.write.option("maxRecordsPerFile", 100000).option(
                "header", "true"
            ).parquet(
                self.config[Constant_Config.STEP_CONFIG]["output_location"],
                mode="append",
                partitionBy=self.config[Constant_Config.STEP_CONFIG]["partition"],
            )
        else:
            output_data.write.mode("overwrite").option("header", "true").parquet(
                self.config[Constant_Config.STEP_CONFIG]["output_location"]
            )
        return None

    @exception_handler
    def get_custom_argparse_configs(self, parser):
        """
        :params parser: the arg parse object to add custom CLI params to
        :return parser:
        """
        parser.add_argument(
            "--input-source",
            action="store",
            type=str,
            required=False,
            help="S3 input file location",
        )
        parser.add_argument(
            "--output-location",
            action="store",
            type=str,
            required=True,
            help="S3 output file location",
        )
        return parser

    @exception_handler
    def add_custom_config_fields(self, config, args):
        """
        Overloading to add custom config fields
        :params config: config object created in get_base_argparse_configs
        :params args: args given to argparse, used in creating custom config pieces
        :return config: new config object with desired custom fields
        """

        config[Constant_Config.STEP_CONFIG]["input_source"] = args["input_source"]
        config[Constant_Config.STEP_CONFIG]["output_location"] = args["output_location"]
        config[Constant_Config.STEP_CONFIG]["metadata"] = {}

        return config

    def parse_fixed_width(self, data_fw):
        """
        Parse fixed width text documents
        Fixed width column lengths and information given in {step}.yaml
        :params input_file_loc: location of input file
        :return data_fw: fixed width text file parsed and transformed into spark df
        """
        for k in self.config[Constant_Config.STEP_CONFIG][
            "input_to_output_cols"
        ].keys():
            # skip metadata columns
            if not re.fullmatch(r".*\{.*\}.*", k):
                assert (
                    self.config[Constant_Config.STEP_CONFIG]["fixed_width_col_lengths"][
                        k
                    ]
                    > 0
                )
        cursor = 1
        assert len(data_fw.columns) == 1
        column_name = data_fw.columns[0]
        for k, v in self.config[Constant_Config.STEP_CONFIG][
            "fixed_width_col_lengths"
        ].items():
            data_fw = data_fw.withColumn(k, data_fw[column_name].substr(cursor, v))
            cursor += v
        data_fw = data_fw.select(
            [
                i
                for i in self.config[Constant_Config.STEP_CONFIG][
                    "fixed_width_col_lengths"
                ].keys()
            ]
        )

        return data_fw

    def add_metadata(self, df):
        """
        Add metadata fields to SSD dataframe according to whether metadata field present in output-loop yaml
        :params df: SSD dataframe created by outputloop
        :params output_loop: yaml config from step yaml file for SSD output loop
        :returns df: SSD dataframe with metadata added
        """
        # Metadata columns: source, job_id, created_at, client
        for k, v in self.config[Constant_Config.STEP_CONFIG][
            "input_to_output_cols"
        ].items():
            if re.fullmatch(r".*\{.*\}.*", k):
                if k == "{SOURCE}":
                    df = df.withColumn(
                        "cad_source_name",
                        F.lit(self.config[Constant_Config.STEP_CONFIG]["input_source"]),
                    )
                if k == "{CLIENT}":
                    df = df.withColumn("cad_client", F.lit(self.config["client"]))
                if k == "{CREATED_AT}":
                    df = df.withColumn("cad_created_at", F.lit(F.current_timestamp()))
                if k == "{JOB_ID}":
                    df = df.withColumn("cad_job_id", F.lit(self.config["job_id"]))
        return df


if __name__ == "__main__":
    text_to_parquet = TextToParquet()
    text_to_parquet.main()



"""
logging
~~~~~~~
This module is for non-spark related module logging. For spark use logging.py
"""
import logging

def logger():
    """
    Creates a logging object and returns it
    """
    logging.basicConfig(
        format='%(asctime)s - %(levelname)s - %(message)s',
        level=logging.INFO)
    return logging


